(*
Copyright 2009-2013 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Matchbounds
imports
  Multihole_Context
  RPO
  Term_Multiset
  Trs_Impl
begin

fun base :: "'f \<times> nat \<Rightarrow> 'f" where "base (f,h) = f"
fun height :: "'f \<times> nat \<Rightarrow> nat" where "height (f,h) = h"
fun lift :: "nat \<Rightarrow> 'f \<Rightarrow> 'f \<times> nat" where "lift h f = (f,h)"
abbreviation base_term where "base_term \<equiv> map_term base"
abbreviation lift_term where "lift_term h \<equiv> map_term (lift h)"

lemma base_lift: "base_term (lift_term h s) = s"
  by (simp only: map_term_comp o_def, simp add: map_term_ident[unfolded id_def])

fun sym_collect :: "(('f,'v)term \<Rightarrow> bool) \<Rightarrow> ('f,'v)term \<Rightarrow> 'f list"
where "sym_collect p (Var x) = []"
  |   "sym_collect p (Fun f ts) = (if p (Fun f ts) then [f] else []) @ 
  concat (map (sym_collect p) ts)"

definition cover :: "(('f,'v)rule \<Rightarrow> ('f,'v)term \<Rightarrow> bool) \<Rightarrow> ('f,'v)trs \<Rightarrow> ('f \<times> nat,'v)trs"
  where "cover ff R \<equiv> { (l', lift_term h r) | l r l' h. 
       (l,r) \<in> R 
     \<and> base_term l' = l 
     \<and> h = Suc (min_list (map height (sym_collect (\<lambda> t'. ff (l,r) (base_term t')) l')))}"

lemma cover_left_linear: assumes "left_linear_trs R"
  shows "left_linear_trs (cover ff R)"
  using assms
  unfolding left_linear_trs_def cover_def
  by (clarify, insert linear_map_term[of base], auto)

lemma cover_var_condition: assumes "\<And> l r. (l,r) \<in> R \<Longrightarrow> vars_term r \<subseteq> vars_term l"
  and "(l,r) \<in> cover ff R"
  shows "vars_term r \<subseteq> vars_term l"
  using assms vars_map_term2[of base]
  unfolding cover_def
  by force

fun roof :: "('f,'v)rule \<Rightarrow> ('f,'v) term \<Rightarrow> bool"
  where [code del]: "roof (l,r) = (\<lambda> t. vars_term r \<subseteq> vars_term t)"


lemma height_lift: assumes "(f,n) \<in> funas_term (lift_term h t)"
  shows "height f = h"
using assms
by (induct t, auto)

definition raise :: "('f \<times> nat,'v) trs" where
  "raise \<equiv> {(Fun (f,n) ts, Fun (f, Suc n) ts) | f n ts. True}"

lemma roof_raise_locally_SN: assumes wf_trs: "wf_trs R" shows "locally_terminating (cover roof (R :: ('f,'v)trs) \<union> raise)"
  unfolding locally_terminating_def
proof (intro allI impI)
  fix F :: "('f \<times> nat)sig"
  assume "finite F"
  from finite_list[OF this] obtain Fs where F: "F = set Fs" by auto
  let ?hs = "map (\<lambda>(f,n). height f) Fs"
  obtain m where m: "m = Suc (max_list ?hs)" by auto
  have small: "\<And> f n. (f,n) \<in> F \<Longrightarrow> height f < m" unfolding F m using max_list[of _ ?hs] by force
  let ?S = "sig_step F (rstep (cover roof R \<union> raise))"
  let ?R = "sig_step F (cover roof R \<union> raise)"
  have subset: "?S \<subseteq> rstep ?R" 
  proof 
    fix s t
    assume "(s,t) \<in> ?S"
    from sig_stepE[OF this] have sF: "funas_term s \<subseteq> F" and tF: "funas_term t \<subseteq> F" and step: "(s,t) \<in> rstep (cover roof R \<union> raise)" by auto
    from step obtain C l r \<sigma> where lr: "(l,r) \<in> cover roof R \<union> raise" 
      and s: "s = C\<langle>l\<cdot>\<sigma>\<rangle>" and t: "t = C\<langle>r \<cdot> \<sigma>\<rangle>" by blast
    from sF tF have F: "funas_term l \<subseteq> F" "funas_term r \<subseteq> F" unfolding s t 
      by (auto simp: funas_term_subst)
    from sig_stepI[OF lr this]
    show "(s,t) \<in> rstep ?R" unfolding s t by auto
  qed
  let ?prc = "\<lambda> (f :: 'f \<times> nat) n g k. (height f < height g \<and> height g \<le> m, height f < height g \<and> height g \<le> m \<or> f = g \<and> n = k)"
  let ?prl = "\<lambda> _. False"
  interpret rpo_pr ?prc ?prl 
  proof(unfold_locales, unfold fst_conv)
    show "SN {((f,n), (g,k)) | f n g k. height f < height g \<and> height g \<le> m}"
      by (rule SN_subset[OF SN_inv_image[of _ "\<lambda>(f,n). height f", unfolded inv_image_def, OF SN_nat_bounded[of m]]], auto)
  qed auto
  show "SN ?S"
  proof (rule SN_subset[OF _ subset], rule rpo_manna_ness, unfold sig_step_def, clarify)
    fix l r
    assume lr: "(l,r) \<in> cover roof R \<union> raise"
    and wfl: "funas_term l \<subseteq> F" and wfr: "funas_term r \<subseteq> F"
    {
      assume lr: "(l,r) \<in> cover roof R"
      from lr[unfolded cover_def]
      obtain ll rr h where base: "base_term l = ll" and llrr: "(ll,rr) \<in> R" and
        r: "r = lift_term h rr" and
        h: "h = Suc (min_list (map height (sym_collect (\<lambda> t'. roof (ll,rr) (base_term t')) l)))" by auto
      from wf_trs[unfolded wf_trs_def] llrr obtain ff lls where ll: "ll = Fun ff lls" and vars: "vars_term rr \<subseteq> vars_term ll" 
        by (cases ll, force+)
      from vars r base vars_map_term2[of "lift h" rr] vars_map_term2[of base l] 
      have vars: "vars_term r \<subseteq> vars_term l" and vars2: "vars_term rr = vars_term r" by auto
      obtain vr where vr: "vars_term r = vr" by auto
      from base ll have "is_Fun l" by (cases l, auto)
      with vars have "\<exists> f' ls'. l \<unrhd> Fun f' ls' \<and> Suc (height f') = h \<and> vars_term r \<subseteq> vars_term (Fun f' ls')" unfolding h roof.simps vars2[symmetric] vars_map_term2 unfolding vars2 vr nat.inject
      proof (induct l)
        case (Var x) thus ?case by simp
      next
        case (Fun f ls) 
        let ?fs = "\<lambda> s. sym_collect (\<lambda> t'. vr \<subseteq> vars_term t') s"
        let ?hs = "\<lambda> s. map height (?fs s)"
        let ?h = "\<lambda> s. min_list (?hs s)"
        show ?case 
        proof (cases "height f = ?h (Fun f ls)")
          case True
          show ?thesis unfolding True[symmetric]
            by (intro exI conjI, rule supteq.refl, rule, rule Fun(2))
        next
          case False        
          let ?fls = "concat (map ?fs ls)"
          from Fun(2) have id: "?fs (Fun f ls) = f # ?fls" by auto
          hence id: "?hs (Fun f ls) = height f # map height ?fls" by auto
          hence hsnNil: "?hs (Fun f ls) \<noteq> []" by simp
          from min_list_ex[OF hsnNil] obtain h where mem: "h \<in> set (?hs (Fun f ls))"
            and min: "h = min_list (?hs (Fun f ls))" by auto
          from False mem min have "h \<in> set (map height ?fls)" unfolding id by auto
          then obtain l where l: "l \<in> set ls" and h: "h \<in> set (map height (?fs l))"
            by auto
          have subset: "set (map height (?fs l)) \<subseteq> set (?hs (Fun f ls))" using l by auto
          from h have vr: "vr \<subseteq> vars_term l" by (induct l, auto)
          from h have var: "is_Fun l" by (cases l, auto)
          from Fun(1)[OF l vr var] obtain f' ls' where supteq: "l \<unrhd> Fun f' ls'"
            and hf': "height f' = min_list (?hs l)" and vr: "vr \<subseteq> vars_term (Fun f' ls')"
            by auto
          from l supteq have supteq: "Fun f ls \<unrhd> Fun f' ls'" by auto
          show ?thesis 
            by (intro exI conjI, rule supteq, unfold hf' min_list_subset[OF subset h[unfolded min]], insert vr, auto)
        qed        
      qed      
      then obtain f' ls' where sub: "l \<unrhd> Fun f' ls'" and h: "h = Suc (height f')" and vars: "vars_term r \<subseteq> vars_term (Fun f' ls')" by auto
      from wfl sub have "(f',length ls') \<in> F" by auto
      from small[OF this] have f'm: "height f' < m" .
      {
        fix g k
        assume "(g,k) \<in> funas_term r"
        from height_lift[OF this[unfolded r]]
        have "height g = h" by simp        
        with h f'm have "height f' < height g \<and> height g \<le> m" by simp
      } note main = this
      have "rpo_pr 0 (Fun f' ls') r = (True,True)" 
        by (rule rpo_large_sym[OF _ vars], insert main, auto)    
      from supteq_imp_rpo_nstri[OF sub] this
      have "fst (rpo_pr 0 l r)" using rpo_compat by force
    }
    moreover
    {
      assume "(l,r) \<in> raise"
      then obtain f n ts where l: "l = Fun (f,n) ts" and r: "r = Fun (f,Suc n) ts" unfolding raise_def by auto
      {
        fix t
        assume t: "t \<in> set ts"
        hence "Fun (f,n) ts \<rhd> t" by auto
        from supt_imp_rpo_stri[OF this] have "fst (rpo_pr 0 (Fun (f,n) ts) t)" .
      } note arg = this
      from wfr[unfolded r] have "((f, Suc n),length ts) \<in> F" by auto
      from small[OF this] have "Suc n < m" by auto
      hence "fst (rpo_pr 0 l r)" unfolding l r rpo.simps Let_def using arg by auto
    }
    ultimately show "fst (rpo_pr 0 l r)" using lr by blast
  qed  
qed

definition match :: "('f,'v)rule \<Rightarrow> ('f,'v)term \<Rightarrow> bool"
  where "match \<equiv> \<lambda> _ _. True"


lemma sym_collect_match: "sym_collect (\<lambda> t'. match lr (f t')) t = funs_term_list' t" unfolding match_def
proof (induct t)
  case (Var x) show ?case by (simp add: funs_term_list'.simps)
next
  case (Fun f ts)
  show ?case 
    by (simp add: funs_term_list'.simps, insert Fun, induct ts, auto simp: funs_term_list'.simps)
qed

    
context
  fixes R :: "('f,'v)trs"
  assumes wf: "wf_trs R"
  and non_duplicating: "\<And> l r. (l,r) \<in> R \<Longrightarrow> vars_term_ms r \<le> vars_term_ms l"
begin

context
  fixes F :: "('f \<times> nat)sig"
  and m :: nat
  and k :: nat
  assumes small: "\<And> f n. (f,n) \<in> F \<Longrightarrow> height f < m"
  and small2: "\<And> l r. (l,r) \<in> R \<Longrightarrow> length (funas_term_list' r) \<le> k"
  and k1: "1 \<le> k"
begin
abbreviation mm where "mm \<equiv> \<lambda> s. funs_term_ms (map_term (\<lambda> f. m - height f) s)"

lemma bound_mset_mm_size: "bound_mset k (mm t) \<le> size t * Suc k^m"
proof -
  {
    fix j
    assume "j \<in> set_of (mm t)"
    hence "j < Suc m"
      by (induct t, auto)
  }
  from bound_mset_linear_in_size[of "mm t" "Suc m" k, OF this] 
  obtain x where "bound_mset k (mm t) = 0 \<or> x < Suc m \<and> bound_mset k (mm t) \<le> size (mm t) * Suc k ^ x" (is "?A \<or> ?B") by auto
  thus ?thesis
  proof
    assume ?A thus ?thesis by simp
  next
    assume ?B
    hence "bound_mset k (mm t) \<le> size (mm t) * Suc k ^ x" by simp
    also have "size (mm t) \<le> size t"
    proof (induct t)
      case (Fun f ts)
      thus ?case by (induct ts, force+)
    qed simp
    also have "Suc k ^ x \<le> Suc k ^ m"
      by (rule power_increasing, insert `?B`, auto)
    finally show ?thesis by auto
  qed
qed

lemma sig_match_raise_step_imp_bmult_less: assumes "(t,s) \<in> sig_step F (rstep (cover match R \<union> raise))"
  shows "(mm s, mm t) \<in> bmult_less k"
  unfolding bmult_less_def
proof -
  let ?lt = "{(x,y :: nat). x < y}"
  let ?le = "\<lambda> s t. (mm s, mm t) \<in> bounded_mult k ?lt"
  from assms
  have ts: "(t,s) \<in> rstep (cover match R \<union> raise)"
    and t: "funas_term t \<subseteq> F"
    and s: "funas_term s \<subseteq> F" by auto      
  thus "?le s t" 
  proof (induct)
    case (IH C \<sigma> l r)        
    let ?C = "funs_ctxt_ms (map_ctxt (\<lambda> f. m - height f) C)"
    let ?\<sigma> = "\<lambda> t. Union_mset (image_mset (\<lambda> x. funs_term_ms ((map_subst (\<lambda>f. m - height f) \<sigma>) x)) (vars_term_ms t))"
    have id: "\<And> t. mm (C\<langle>t \<cdot> \<sigma>\<rangle>) = ?C + mm t + ?\<sigma> t"
      by (simp  add: funs_ms_ctxt_apply funs_ms_subst_apply multiset_eq_iff)
    from IH(1) show "?le (C\<langle>r \<cdot> \<sigma>\<rangle>) (C\<langle>l \<cdot> \<sigma>\<rangle>)"        
    proof
      assume lr: "(l,r) \<in> cover match R"
      from lr[unfolded cover_def] obtain ll rr h where base: "base_term l = ll" and llrr: "(ll,rr) \<in> R" and
        r: "r = lift_term h rr" and
        h: "h = Suc (min_list (map height (sym_collect (\<lambda> t'. match (ll,rr) (base_term t')) l)))" by auto
      have "size (mm r) = length (funas_term_list' rr)" unfolding r
      proof (induct rr)
        case (Fun f ss)
        thus ?case by (induct ss, auto simp: funas_term_list'.simps)
      qed (simp add: funas_term_list'.simps)
      also have "\<dots> \<le> k" by (rule small2[OF llrr])
      finally have k: "size (mm r) \<le> k" .
      let ?h = "min_list (map height (funs_term_list' l))"
      from h[unfolded sym_collect_match]
      have hS: "h = Suc (?h)" by simp
      from wf[unfolded wf_trs_def] llrr obtain ff lls where ll: "ll = Fun ff lls" by (cases ll, force+)
      {
        let ?mmm = "\<lambda>ls. multiset_of (map (\<lambda>f. m - height f) ls)"
        let ?fl = "funs_term_list' l"
        let ?mm = "min_list (map height ?fl)"
        from base ll have ne: "map height ?fl \<noteq> []" by (cases l, auto simp: funs_term_list'.simps)
        from min_list_ex[OF this] obtain f where
          f: "f \<in> set ?fl" and hf: "?mm = height f" by auto
        from f[unfolded funs_term_list'_sound funs_term_funas_term] 
        obtain n where fn: "(f,n) \<in> funas_term l" by auto
        from IH(2) have l: "funas_term l \<subseteq> F" by (auto simp: funas_term_subst)
        from fn l have "(f,n) \<in> F" by auto
        from small[OF this] have "height f < m" .
        hence m: "?h < m" unfolding hf by auto
        from f[unfolded set_conv_nth] obtain i where i: "i < length ?fl" 
          and f: "?fl ! i = f" by auto
        from id_take_nth_drop[OF i, unfolded f] obtain bef aft 
          where fl: "?fl = bef @ f # aft" by auto        
        have "\<exists> L. mm l = L + {# m - ?h #}" unfolding multiset_of_funs_term_list'[symmetric] 
          unfolding hf funs_term_list'_map_term
          unfolding fl
          by (rule exI[of _ "?mmm bef + ?mmm aft"], simp add: multiset_eq_iff)
        note this and m
      } note Lm = this
      then obtain L where L: "mm l = L + {# m - ?h #}" by auto
      {
        fix h'
        assume h': "h' \<in># mm r" 
        from this[unfolded multiset_of_funs_term_list'[symmetric]]
        have "h' \<in> (\<lambda>f. m - height f) ` (funs_term r)" 
          unfolding funs_term_list'_map_term in_multiset_in_set using funs_term_list'_sound by auto
        from this[unfolded funs_term_funas_term]  
        obtain f n where f: "(f,n) \<in> funas_term r" and h': "h' = m - height f"
          by auto
        from height_lift[OF f[unfolded r]] h' have h': "h' = m - h" by auto
        have "h' < m - ?h" unfolding h' hS using Lm(2) by auto
      }
      with L obtain h where l: "mm l = L + {#h#}" and h: "\<forall>h'. h' \<in># mm r \<longrightarrow> h' < h" by auto
      from non_duplicating[OF llrr] have "vars_term_ms (base_term r) <= vars_term_ms (base_term l)" unfolding base r by auto 
      hence "vars_term_ms r <= vars_term_ms l" by simp
      then obtain M where Mid: "vars_term_ms l = vars_term_ms r + M" 
        unfolding mset_le_exists_conv ..
      obtain \<sigma>diff where \<sigma>l: "?\<sigma> l = ?\<sigma> r + \<sigma>diff" unfolding Mid
        by auto
      have l: "mm (C\<langle>l \<cdot> \<sigma>\<rangle>) = (?C + ?\<sigma> r) + (L + {#h#} + \<sigma>diff)"
        unfolding id l \<sigma>l multiset_eq_iff by auto
      have r: "mm (C\<langle>r \<cdot> \<sigma>\<rangle>) = (?C + ?\<sigma> r) + mm r"
        unfolding id multiset_eq_iff by auto
      show "?le (C\<langle>r \<cdot> \<sigma>\<rangle>) (C\<langle>l \<cdot> \<sigma>\<rangle>)"        
        unfolding l r          
        by (rule bounded_mult, insert k h, auto)
    next
      assume lr: "(l,r) \<in> raise"
      then obtain f n ts where l: "l = Fun (f,n) ts" and r: "r = Fun (f,Suc n) ts" unfolding raise_def by auto
      from IH(3)[unfolded r] have "((f, Suc n), length ts) \<in> F" by auto
      from small[OF this] have "Suc n < m" by auto
      hence small: "m - Suc n < m - n" by auto
      have "\<exists> L. mm (l \<cdot> \<sigma>) = L + {# m - n #} \<and> mm (r \<cdot> \<sigma>) = L + {# m - Suc n #}"
        unfolding l r subst_apply_term.simps
        unfolding funs_term_ms_map_term by auto
      then obtain L where id: "mm (l \<cdot> \<sigma>) = L + {# m - n #}" "mm (r \<cdot> \<sigma>) = L + {# m - Suc n #}" by auto
      have id: "mm (C \<langle>l \<cdot> \<sigma> \<rangle>) = (?C + L) + {# m - n #}" "mm (C \<langle>r \<cdot> \<sigma> \<rangle>) = (?C + L) + {# m - Suc n #}" 
        unfolding funs_ms_ctxt_apply map_ctxt_distrib id multiset_eq_iff by auto
      show ?thesis unfolding id 
        by (rule bounded_mult, insert small k1, auto)
    qed
  qed
qed

lemma sig_match_raise_step_imp_mult_step: assumes "(t,s) \<in> sig_step F (rstep (cover match R \<union> raise))"
  shows "(mm s, mm t) \<in> mult {(x, y). x < y}"
  by (rule bounded_mult_into_mult[OF _ sig_match_raise_step_imp_bmult_less[OF assms, unfolded bmult_less_def]], auto simp: trans_def)

lemma sig_match_union_raise_linear_bound: assumes steps: "(s,t) \<in> (sig_step F (rstep (cover match R \<union> raise)))^^n" (is "_ \<in> ?R ^^ n")
  shows "n \<le> size s * Suc k ^ m"
proof -
  from steps have "(mm t, mm s) \<in> (bmult_less k)^^n"
  proof (induct n arbitrary: t)
    case (Suc n)
    from Suc(2) obtain u where su: "(s,u) \<in> ?R ^^ n" and ut: "(u,t) \<in> ?R" by auto
    from sig_match_raise_step_imp_bmult_less[OF ut] Suc(1)[OF su] show ?case by (rule relpow_Suc_I2)
  qed simp
  from bound_mset_bmult_less[OF this] 
  have "n \<le> bound_mset k (mm s)" .
  also have "\<dots> \<le> size s * Suc k ^ m"
    by (rule bound_mset_mm_size)
  finally show ?thesis .
qed  

lemma SN_sig_step_match_raise: "SN (sig_step F (rstep (cover Matchbounds.match R \<union> raise)))"
proof -
  let ?gt = "\<lambda> s t. (s,t) \<in> rstep (cover match R \<union> raise) \<and> funas_term s \<subseteq> F \<and> funas_term t \<subseteq> F"  
  let ?lt = "{(x,y :: nat). x < y}"
  let ?le = "\<lambda> s t. (mm s, mm t) \<in> mult ?lt"
  let ?R = "{(s,t). ?le s t}"
  have id: "{(s,t). ?gt t s} = (sig_step F (rstep (cover match R \<union> raise)))^-1" 
    unfolding sig_step_def by auto
  have trans: "trans ?lt" unfolding trans_def by auto
  have main: "wf {(s,t). ?gt t s}" 
  proof (rule wf_subset)
    show "wf ?R"
      using wf_inv_image[OF wf_mult[OF wf_less], of mm]
      unfolding inv_image_def .
  next    
    show "{(s,t). ?gt t s} \<subseteq> ?R" using sig_match_raise_step_imp_mult_step
      unfolding id by auto
  qed
  show "SN (sig_step F (rstep (cover match R \<union> raise)))"
    by (rule wf_imp_SN[OF main[unfolded id]])
qed
end

lemma match_raise_locally_SN: assumes finR: "finite R" 
  shows "locally_terminating (cover match R \<union> raise)"
  unfolding locally_terminating_def
proof (intro allI impI)
  fix F :: "('f \<times> nat) sig" 
  assume fin: "finite F" 
  from finite_list[OF fin] obtain Fs where F: "F = set Fs" by auto
  let ?hs = "map (\<lambda>(f,n). height f) Fs"
  def m \<equiv> "Suc (max_list ?hs)"
  have small: "\<And> f n. (f,n) \<in> F \<Longrightarrow> height f < m" unfolding F m_def using max_list[of _ ?hs] by force
  from finite_list[OF finR] obtain Rs where R: "R = set Rs" by auto
  let ?fs = "map (\<lambda> (l,r). length (funas_term_list' r)) Rs"
  def k \<equiv> "Suc (max_list ?fs)"
  have small2: "\<And> l r. (l,r) \<in> R \<Longrightarrow> length (funas_term_list' r) \<le> k" unfolding R k_def using max_list[of _ ?fs] by force
  have k1: "1 \<le> k" unfolding k_def by simp
  show "SN (sig_step F (rstep (cover Matchbounds.match R \<union> raise)))"
    by (rule SN_sig_step_match_raise[OF small small2 k1])
qed 
end
hide_const mm

definition ge_raise :: "('f \<times> nat) sig \<Rightarrow> (('f \<times> nat),'v)term \<Rightarrow> (('f \<times> nat),'v)term \<Rightarrow> bool" where
  "ge_raise F s t \<equiv> (t,s) \<in> (sig_step F (rstep raise))^*"

lemma ge_raise_trans: "ge_raise F s t \<Longrightarrow> ge_raise F t u \<Longrightarrow> ge_raise F s u" unfolding ge_raise_def by auto

fun max_raise :: "(('f \<times> nat),'v)rule \<Rightarrow> (('f \<times> nat),'v)term option" where
  "max_raise (Var x,Var y) = (do {
     guard (x = y);
     Some (Var y)
   })"
| "max_raise (Fun (f1,h1) ts1, Fun (f2,h2) ts2) = (do {
    guard ((f1,length ts1) = (f2,length ts2));
    ts \<leftarrow> mapM max_raise (zip ts1 ts2);
    Some (Fun (f2,max h1 h2) ts)
  })"
| "max_raise _ = None"

lemma max_raise_refl: "max_raise (s,s) = Some s"
proof (induct s)
  case (Fun f ss)
  have "\<forall>x\<in>set (zip ss ss). max_raise x \<noteq> None" using Fun unfolding set_conv_nth by force
  hence "mapM max_raise (zip ss ss) = Some (map (\<lambda>x. the (max_raise x)) (zip ss ss))" unfolding mapM_map by auto
  also have "map (\<lambda>x. the (max_raise x)) (zip ss ss) = ss" 
    unfolding list_eq_iff_nth_eq using Fun[unfolded set_conv_nth]
    by force
  finally show ?case by (cases f, auto)
qed auto

lemma max_raise_sym: "max_raise (s,t) = max_raise (t,s)"
proof (induct s arbitrary: t)
  case (Var x t)
  show ?case by (cases t, auto simp: guard_def)
next
  case (Fun f ss t) note IH = this
  obtain fn fh where f: "f = (fn, fh)" by force
  show ?case 
  proof (cases t)
    case (Fun g ts)
    obtain gn gh where g: "g = (gn, gh)" by force
    let ?f = "(fn,length ss)"
    let ?g = "(gn,length ts)"
    have "(?f = ?g) = (?g = ?f)" "max fh gh = max gh fh" by auto
    note id = Fun f g this
    show ?thesis 
    proof (cases "?g = ?f")
      case False
      thus ?thesis unfolding max_raise.simps guard_def id by auto
    next
      case True
      hence len: "length ss = length ts" by auto
      let ?ssts = "mapM max_raise (zip ss ts)"
      let ?tsss = "mapM max_raise (zip ts ss)"
      from True have "?thesis = (?ssts = ?tsss)"
        unfolding max_raise.simps id guard_def 
        by (simp, cases ?ssts, (cases ?tsss, (auto)[2])+)
      also have "..." using len IH
        by (induct ss ts rule: list_induct2, force+)
      finally show ?thesis by simp
    qed
  qed auto
qed


lemma max_raise_base_Some: "base_term s = base_term t \<Longrightarrow> (\<exists> u. max_raise (s,t) = Some u \<and> base_term u = base_term t)"
proof (induct s arbitrary: t)
  case (Var x)
  hence t: "t = Var x" by (cases t, auto)
  thus ?case by auto
next
  case (Fun fp ss)
  then obtain gp ts where t: "t = Fun gp ts" by (cases t, auto)
  obtain f fh where fp: "fp = (f,fh)" by force
  note cond = Fun(2)[unfolded t fp]
  from cond obtain gh where gp: "gp = (f,gh)" by (cases gp, auto)
  from cond have id: "map base_term ss = map base_term ts" by auto
  from arg_cong[OF this, of length] have len: "length ss = length ts" by auto
  let ?p = "\<lambda> i u. max_raise (ss ! i, ts ! i) = Some u \<and> base_term u = base_term (ts ! i)"
  {
    fix i
    assume i: "i < length ts"
    hence "ss ! i \<in> set ss" using len by auto
    from Fun(1)[OF this map_nth_conv[OF id, rule_format]] i len
    have "\<exists> u. ?p i u" by auto
  }
  hence "\<forall> i. \<exists> u. i < length ts \<longrightarrow> ?p i u" by blast
  from choice[OF this] obtain u where IH: "\<And> i. i < length ts \<Longrightarrow> ?p i (u i)" by blast
  let ?u = "Fun (f,max fh gh) (map u [0 ..< length ts])"
  have id1: "map (base_term \<circ> u) [0..<length ts] = map base_term ts"
    by  (rule nth_map_conv, insert IH, auto)
  have id2: "mapM max_raise (zip ss ts) = Some (map u [0..<length ts])"
    unfolding mapM_map set_zip using IH by (auto intro!: nth_map_conv simp: len)
  show ?case unfolding t fp gp
    by (rule exI[of _ ?u], auto simp: len id1 id2)
qed
  
  
lemma max_raise_ge_raise: assumes F: "\<And> f h h' n. ((f,h),n) \<in> F \<Longrightarrow> h' \<le> h \<Longrightarrow> ((f,h'),n) \<in> F"
  shows "max_raise (s,t) = Some u \<Longrightarrow> funas_term s \<subseteq> F \<Longrightarrow> funas_term t \<subseteq> F \<Longrightarrow> ge_raise F u s \<and> funas_term u \<subseteq> F" unfolding ge_raise_def
proof (induct s arbitrary: t u)
  case (Var x t u)
  thus ?case by (cases t, auto simp: guard_def split: if_splits)
next
  case (Fun f ss t u) 
  note IH = Fun(1)
  note some = Fun(2)
  obtain fn fh where f: "f = (fn, fh)" by force
  from some obtain g ts where t: "t = Fun g ts" by (cases t, auto)
  obtain gn gh where g: "g = (gn, gh)" by force
  note some = some[unfolded t f g, simplified]
  from some have fn: "gn = fn" and len: "length ts = length ss" by auto
  from some obtain us where map: "mapM max_raise (zip ss ts) = Some us"
    and u: "u = Fun ((fn,max fh gh)) us" by auto
  note map = map[unfolded mapM_map]
  from map have us: "us = map (\<lambda>x. the (max_raise x)) (zip ss ts)" by (auto split: if_splits)
  from us len have len2: "length ss = length us" by auto
  have "\<exists> k. max fh gh = fh + k" by presburger
  then obtain k where max: "max fh gh = fh + k" by auto
  from Fun(3)[unfolded f g] len2 have F1: "((fn,fh),length us) \<in> F" "UNION (set ss) funas_term \<subseteq> F" by auto
  from Fun(4)[unfolded t f g fn] len len2 have F2: "((fn,gh),length us) \<in> F" "UNION (set ts) funas_term \<subseteq> F" by auto
  let ?R = "sig_step F (rstep raise)"  
  let ?u = "\<lambda> h. Fun (fn,fh + h) us"
  {
    fix i
    assume i: "i < length ss"
    hence mem: "ss ! i \<in> set ss" by auto
    from i have "(ss ! i, ts ! i) \<in> set (zip ss ts)" using len 
      unfolding set_zip by auto
    with map have max: "max_raise (ss ! i, ts ! i) \<noteq> None" by (auto split: if_splits)
    have "(ss ! i, us ! i) \<in> ?R^* \<and> funas_term (us ! i) \<subseteq> F"
    proof (rule IH[OF mem])
      show "max_raise (ss ! i, ts ! i) = Some (us ! i)"
        unfolding us using max i len len2 by auto
    qed (insert F1(2) F2(2) i len len2, unfold set_conv_nth, auto, blast)
  } note IH = this
  have F3: "UNION (set us) funas_term \<subseteq> F" unfolding set_conv_nth using IH[unfolded len2] by auto
  from F1(1) F2(1) have "((fn, max fh gh), length us) \<in> F" unfolding max_def by (auto split: if_splits)
  hence Fk: "((fn,fh + k),length us) \<in> F" unfolding max by auto
  have "(Fun f ss, Fun f us) \<in> ?R^*"
  proof (rule all_ctxt_closedE[OF all_ctxt_closed_sig_rsteps _ len2], unfold f len)
    show "((fn, fh), length us) \<in> F" by fact
    fix i
    assume i: "i < length ss"
    from IH[OF this] have "(ss ! i, us ! i) \<in> ?R^*" "funas_term (us ! i) \<subseteq> F" by auto
    show "(ss ! i, us ! i) \<in> ?R^*" by fact
    show "funas_term (us ! i) \<subseteq> F" by fact
    show "funas_term (ss ! i) \<subseteq> F" using F1(2) i unfolding set_conv_nth by auto
  qed
  also have "(Fun f us, u) \<in> ?R^*"
    unfolding u f fn max using Fk
  proof (induct k)
    case (Suc k)
    note k = Suc(2)
    have sk: "((fn,fh + k),length us) \<in> F" by (rule F[OF k], simp)
    have "(Fun (fn,fh) us, ?u k) \<in> ?R^*" by (rule Suc(1)[OF sk])
    also have "(?u k, ?u (Suc k)) \<in> ?R"
      by (rule sig_stepI[OF rstepI[of "?u k" "?u (Suc k)" _ _ Hole Var]], unfold raise_def, insert F3 k sk, auto)
    finally show ?case .
  qed auto
  finally show ?case unfolding u max using Fk F3 by auto
qed

fun max_raise_list :: "('f \<times> nat,'v)term list \<Rightarrow> ('f \<times> nat,'v) term option" where
  "max_raise_list [] = Some undefined"
| "max_raise_list [t] = Some t"
| "max_raise_list (t # s # ts) = do {
     h \<leftarrow> max_raise (t,s);
     max_raise_list (h # ts)
  }"

lemma max_raise_list_ge_raise: assumes F: "\<And> f h h' n. ((f,h),n) \<in> F \<Longrightarrow> h' \<le> h \<Longrightarrow> ((f,h'),n) \<in> F"
  shows "t \<in> set ts \<Longrightarrow> max_raise_list ts = Some s \<Longrightarrow> UNION (set ts) funas_term \<subseteq> F \<Longrightarrow> ge_raise F s t \<and> funas_term s \<subseteq> F"
proof (induct ts arbitrary: t rule: max_raise_list.induct)
  case (3 u v ts t)
  from 3(3) obtain w where w: "max_raise (u,v) = Some w" by auto
  from 3(4) have Fu: "funas_term u \<subseteq> F" and Fv: "funas_term v \<subseteq> F" and Fts: "UNION (set ts) funas_term \<subseteq> F" by auto
  from 3(3) w have "max_raise_list (w # ts) = Some s" by auto
  note IH = 3(1)[OF w _ this]
  from max_raise_ge_raise[OF F w Fu Fv] have Fw: "funas_term w \<subseteq> F" by auto
  with Fts have "UNION (set (w # ts)) funas_term \<subseteq> F" by auto
  note IH = IH[OF _ this]
  from 3(2) have "t \<in> {u,v} \<or> t \<in> set ts" by auto
  thus ?case
  proof
    assume "t \<in> {u,v}"
    with max_raise_ge_raise[OF F w Fu Fv] max_raise_ge_raise[OF F w[unfolded max_raise_sym[of u]] Fv Fu] 
    have ge: "ge_raise F w t" by auto
    from ge_raise_trans[OF IH[THEN conjunct1] ge] IH show ?thesis by auto
  qed (insert IH, auto)
next
  case (2 t u)
  hence u: "u = s" and F: "funas_term s \<subseteq> F" by auto
  show ?case unfolding u ge_raise_def using F by auto
qed auto
  
lemma max_raise_list_base_Some: "base_term ` set ts = {b} \<Longrightarrow> \<exists> u. max_raise_list ts = Some u \<and> base_term u = b"
proof (induct ts rule: max_raise_list.induct)
  case (3 t v vs)
  from 3(2) have "base_term t = base_term v" "base_term v = b" by auto
  from max_raise_base_Some[OF this(1)] this(2) obtain u 
    where u: "max_raise (t,v) = Some u" and bu: "base_term u = b"
    by auto
  from 3(1)[OF u] bu 3(2) obtain v where "max_raise_list (u # vs) = Some v" and "base_term v = b" by auto
  with u show ?case by auto
qed auto

inductive_set raise_step :: "('f \<times> nat,'v)trs \<Rightarrow> ('f \<times> nat,'v)trs" for R :: "('f \<times> nat, 'v)trs" where
  raise_step: "(l,r) \<in> R 
  \<Longrightarrow> split_vars l = (C,xs) 
  \<Longrightarrow> length ss = length xs 
  \<Longrightarrow> (\<And> i j. i < length xs \<Longrightarrow> j < length xs \<Longrightarrow> xs ! i = xs ! j \<Longrightarrow> base_term (ss ! i) = base_term (ss ! j))
  \<Longrightarrow> (\<And> x. \<Theta> x = (if x \<in> vars_term l then the (max_raise_list (map fst (filter (\<lambda> (si,xi). xi = x) (zip ss xs)))) else Var x))
  \<Longrightarrow> (D \<langle> fill_holes C ss \<rangle>, D \<langle> r \<cdot> \<Theta> \<rangle>) \<in> raise_step R"

lemma sig_raise_step_imp_sig_steps:  assumes F_cond: "\<And> f h h' n. ((f,h),n) \<in> F \<Longrightarrow> h' \<le> h \<Longrightarrow> ((f,h'),n) \<in> F" 
  and step: "(s,t) \<in> sig_step F (raise_step R)"
  shows "(s,t) \<in> (sig_step F (rstep raise))^* O (sig_step F (rstep R))"
proof -
  let ?R = "sig_step F (rstep R)"
  let ?Rai = "sig_step F (rstep raise)"
  from step have "(s,t) \<in> raise_step R" "funas_term s \<subseteq> F" "funas_term t \<subseteq> F" by auto
  thus ?thesis
  proof (induct rule: raise_step.induct)
    case (raise_step l r C xs ss \<Theta> D)
    note Theta = raise_step(5)
    note split = raise_step(2)
    note len_ss = raise_step(3)
    note F = raise_step(6,7)  
    from F(1) have FCss: "funas_term (fill_holes C ss) \<subseteq> F" by auto    
    from split_vars_eqf_subst[of l \<Theta>, unfolded split]
    have eqf: "l \<cdot> \<Theta> =f (C, map \<Theta> xs)" by auto
    from eqfE[OF eqf] have ltheta: "l \<cdot> \<Theta> = fill_holes C (map \<Theta> xs)" and nh: "num_holes C = length xs" by auto
    from nh len_ss have "num_holes C = length ss" by simp
    hence Css: "fill_holes C ss =f (C,ss)" by auto
    from FCss[unfolded eqf_funas_term[OF this]] have FC: "funas_mctxt C \<subseteq> F" and Fss: "UNION (set ss) funas_term \<subseteq> F" by auto
    {
      fix i
      assume i: "i < length xs"
      hence id: "map \<Theta> xs ! i = \<Theta> (xs ! i)" by auto
      let ?ts = "map fst [(si, xi)\<leftarrow>zip ss xs . xi = xs ! i]"
      have "(ss ! i, map \<Theta> xs ! i) \<in> (sig_step F (rstep raise))^* \<and> funas_term (map \<Theta> xs ! i) \<subseteq> F" unfolding id 
      proof (rule max_raise_list_ge_raise[unfolded ge_raise_def raise_step, OF F_cond])
        show ss: "ss ! i \<in> set ?ts" using i unfolding set_map set_filter set_zip len_ss by force
        have "base_term ` set ?ts = {base_term (ss ! i)}" (is "?l = ?r")
        proof
          show "?r \<subseteq> ?l" using ss by auto
          show "?l \<subseteq> ?r"
          proof
            fix b
            assume "b \<in> ?l"
            then obtain j where j: "j < length xs" and id: "xs ! j = xs ! i" and b: "b = base_term (ss ! j)"
              unfolding set_map set_filter set_zip len_ss by auto
            from raise_step(4)[OF j i id] show "b \<in> ?r" unfolding b by auto
          qed
        qed                
        from max_raise_list_base_Some[OF this]
        show "max_raise_list ?ts = Some (\<Theta> (xs ! i))" unfolding Theta
          using split_vars_vars_term[of l, symmetric] i unfolding split by auto
      next
        have "set ?ts \<subseteq> set ss" unfolding set_map set_filter set_zip by force
        with Fss         
        show "UNION (set ?ts) funas_term \<subseteq> F" by auto
      qed 
    }
    note main = this
    have FTheta: "UNION (set (map \<Theta> xs)) funas_term \<subseteq> F" unfolding set_map set_conv_nth length_map using main[THEN conjunct2] by force
    have raise: "(fill_holes C ss, fill_holes C (map \<Theta> xs)) \<in> ?Rai^* \<and> fill_holes C (map \<Theta> xs) =f (C, map \<Theta> xs)"
      by (rule eqf_all_ctxt_closed_step[OF all_ctxt_closed_sig_rsteps Css _ _ FCss], unfold length_map, insert main FTheta len_ss, auto)
    from raise ltheta have steps: "(fill_holes C ss, l \<cdot> \<Theta>) \<in> ?Rai^*" by auto
    from eqf_funas_term[OF eqf] FC FTheta have Flt: "funas_term (l \<cdot> \<Theta>) \<subseteq> F" by auto
    from F have FD: "funas_ctxt D \<subseteq> F" by auto
    have steps: "(D\<langle>fill_holes C ss\<rangle>, D\<langle>l \<cdot> \<Theta>\<rangle>) \<in> ?Rai^*" 
      by (rule all_ctxt_closed_ctxtE[OF _ FCss Flt steps FD], auto)
    moreover
    {
      from raise_step(1) have step: "(D\<langle>l \<cdot> \<Theta>\<rangle>, D\<langle>r \<cdot> \<Theta>\<rangle>) \<in> rstep R" by auto
      have "(D\<langle>l \<cdot> \<Theta>\<rangle>, D\<langle>r \<cdot> \<Theta>\<rangle>) \<in> ?R"
        by (rule sig_stepI[OF step _ F(2)], insert Flt FD, auto)
    }
    ultimately show ?case by blast
  qed
qed
    
(* lemma 12 in IC 09 *)
lemma rstep_imp_raise_step: assumes var_cond: "\<And> l r. (l,r) \<in> R \<Longrightarrow> vars_term r \<subseteq> vars_term l"
  shows "(s,t) \<in> rstep R \<Longrightarrow>
  base_term s' = s \<longrightarrow> (\<exists> t'. (s',t') \<in> raise_step (cover ff R) \<and> base_term t' = t)"
proof (induct arbitrary: s' rule: rstep_induct)
  case (IH D \<sigma> l r)
  show ?case
  proof 
    fix s'
    let ?R = "cover ff R"
    assume "base_term s' = D \<langle> l \<cdot> \<sigma> \<rangle>"
    from map_term_ctxt_decomp[OF this] 
    obtain E u where D: "D = map_ctxt base E" and u: "l \<cdot> \<sigma> = base_term u" and s': "s' = E \<langle> u \<rangle>" by blast
    from subst_eq_map_decomp[OF u]
    obtain C xs \<delta>s where u: "u =f (C, \<delta>s)" 
      and split: "split_vars l = (map_mctxt base C, xs)" 
      and \<sigma>: "\<And> i. i < length xs \<Longrightarrow> \<sigma> (xs ! i) = base_term (\<delta>s ! i)" by auto
    from split_vars_map_mctxt[OF split] have splitv: "split_vars (fill_holes C (map Var xs)) = (C,xs)" by auto
    from split_vars_num_holes[of "fill_holes C (map Var xs)", unfolded splitv] have nh: "num_holes C = length (map Var xs)" by simp
    from eqfE[OF u] have u: "u = fill_holes C \<delta>s" and \<delta>s: "num_holes C = length \<delta>s" by auto
    from split_vars_eqf_subst[of l Var, unfolded split] have l: "l =f (map_mctxt base C, map Var xs)" by simp
    have "\<exists> y. (fill_holes C (map Var xs),lift_term y r) \<in> ?R" unfolding cover_def
    proof (rule exI, rule, intro exI, rule conjI[OF refl conjI[OF IH conjI[OF _ refl]]]) 
      show "base_term (fill_holes C (map Var xs)) = l" 
      unfolding eqfE(1)[OF map_term_fill_holes[OF nh]]
      unfolding eqfE[OF l] map_map o_def by auto
    qed
    then obtain n where rule: "(fill_holes C (map Var xs), lift_term n r) \<in> ?R" ..
    from \<delta>s nh have len: "length \<delta>s = length xs" by auto
    def \<Theta> \<equiv> "\<lambda> x. (if x \<in> vars_term (fill_holes C (map Var xs))
         then the (max_raise_list (map fst [(si, xi)\<leftarrow>zip \<delta>s xs . xi = x])) else Var x)"
    show "\<exists>t'. (s', t') \<in> raise_step ?R \<and> base_term t' = D\<langle>r \<cdot> \<sigma>\<rangle>"
    proof (intro exI conjI, unfold s' D u, rule raise_step[OF rule splitv])
      fix i j
      assume i: "i < length xs" and j: "j < length xs" and id: "xs ! i = xs ! j"
      show "base_term (\<delta>s ! i) = base_term (\<delta>s ! j)" 
        using \<sigma>[OF i, symmetric, unfolded id, unfolded \<sigma>[OF j]] i j by auto
    next
      have "(base_term E\<langle>lift_term n r \<cdot> \<Theta>\<rangle> = (map_ctxt base E)\<langle>r \<cdot> \<sigma>\<rangle>) = (r \<cdot> (\<lambda>x. base_term (\<Theta> x)) = r \<cdot> \<sigma>)" (is "?goal = _") 
        by (simp add: base_lift)
      also have "\<dots>"
      proof (rule term_subst_eq)
        fix x
        assume "x \<in> vars_term r"
        with var_cond[OF IH] have x: "x \<in> vars_term l" by auto
        hence xs: "x \<in> set xs" using split_vars_vars_term[of l, unfolded split] by auto
        also have "set xs = vars_term (fill_holes C (map Var xs))" using split_vars_vars_term[of "fill_holes C (map Var xs)", unfolded splitv] by auto
        finally have x: "(x \<in> vars_term (fill_holes C (map Var xs))) = True" by simp
        let ?ts = "map fst [(si, xi)\<leftarrow>zip \<delta>s xs . xi = x]"
        {
          from xs obtain i where i: "i < length xs" and x: "x = xs ! i" unfolding set_conv_nth by auto
          hence "\<delta>s ! i \<in> set ?ts" unfolding set_zip set_map set_filter len by force
          with \<sigma>[OF i] have "\<sigma> x \<in> base_term ` set ?ts" unfolding x by auto
        }
        moreover
        {
          fix t
          assume "t \<in> base_term ` set ?ts"
          then obtain i where t: "t = base_term (\<delta>s ! i)" and i: "i < length xs" and x: "xs ! i = x"
            unfolding set_map set_filter set_zip by force
          from \<sigma>[OF i] t x have "t = \<sigma> x" by auto
        }
        ultimately have "base_term ` set ?ts = {\<sigma> x}" by blast        
        from max_raise_list_base_Some[OF this] obtain u where res: "max_raise_list ?ts = Some u" and u: "base_term u = \<sigma> x" by auto
        have "base_term (\<Theta> x) = base_term (the (max_raise_list ?ts))"
          unfolding \<Theta>_def x by auto
        also have "\<dots> = \<sigma> x" unfolding res using u by auto
        finally show "base_term (\<Theta> x) = \<sigma> x" .
      qed
      finally show ?goal by blast
    qed (auto simp: \<Theta>_def len)
  qed
qed

lemma max_raise_map_vars: fixes vw defines "av \<equiv> map_vars vw"
  assumes g: "ground s" "ground t"
  and r: "max_raise (av s, av t) = Some u"
  shows "\<exists> v. max_raise (s,t) = Some v \<and> ground v \<and> u = av v" 
  using g r
proof (induct s arbitrary: t u)
  case (Fun fl ss ftt u)
  obtain f l where fl: "fl = (f,l)" by force
  from Fun(3) obtain fl' ts where ftt: "ftt = Fun fl' ts" by (cases ftt, auto)
  obtain g l' where fl': "fl' = (g,l')" by force
  note res = Fun(4)[unfolded fl ftt fl' av_def, simplified, folded av_def]
  from res have gf: "g = f" and len: "length ss = length ts" by auto
  def z \<equiv> "zip (map av ss) (map av ts)"
  from len have lenz: "length z = length ts" unfolding z_def by simp
  from res obtain us where Some: "mapM max_raise z = Some us"
    and u: "u = Fun (g, max l l') us" unfolding z_def by auto
  from mapM_Some[OF Some] have lenus: "length us = length ts" using lenz by simp
  let ?p = "\<lambda> i v. max_raise (ss ! i, ts ! i) = Some v \<and> us ! i = av v \<and> ground v \<and> max_raise (z ! i) = Some (av v)"
  {
    fix i
    assume i: "i < length ts"
    with lenz have "i < length z" by simp
    from mapM_Some_idx[OF Some this] obtain y where res: "max_raise (z ! i) = Some y" and us: "us ! i = y"
      by auto
    from i len have mem: "ss ! i \<in> set ss" and "ts ! i \<in> set ts" by auto
    with Fun(2-3)[unfolded ftt] have g: "ground (ss ! i)" "ground (ts ! i)" by auto
    from i len have "z ! i = (av (ss ! i), av (ts ! i))" unfolding z_def by auto
    from Fun(1)[OF mem g res[unfolded this]] res us have 
      "\<exists> v. ?p i v" by auto
  }
  hence "\<forall> i. \<exists> v. i < length ts \<longrightarrow> ?p i v" by blast
  from choice[OF this] obtain vs where IH: "\<And> i. i < length ts \<Longrightarrow> ?p i (vs i)" by blast
  let ?vs = "map vs [0 ..< length ts]"
  have us: "us = map (av \<circ> vs) [0..<length ts]"
    by (rule nth_equalityI, insert len lenus IH, auto)
  have "mapM max_raise (zip ss ts) = Some (map (\<lambda>x. the (max_raise x)) (zip ss ts))" 
    unfolding mapM_map set_zip len using IH by auto
  also have "map (\<lambda>x. the (max_raise x)) (zip ss ts) = ?vs"
    by (rule nth_equalityI, insert IH, auto simp: len)
  finally show ?case unfolding fl ftt fl' gf u
    by (intro exI[of _ "Fun (f, max l l') ?vs"], insert IH, simp add: len av_def us)
qed simp

lemma max_raise_list_map_vars: fixes vw defines "av \<equiv> map_vars vw"
  assumes g: "\<And> s. s \<in> set sss \<Longrightarrow> ground s"
  and r: "max_raise_list (map av sss) = Some u"
  and s: "set sss \<noteq> {}"
  shows "\<exists> v. max_raise_list sss = Some v \<and> u = av v"
  using g r s
proof (induct sss rule: max_raise_list.induct)
  case (3 s t ss)
  from 3(2) have gs: "ground s" and gt: "ground t" by auto
  note r = 3(3)
  from r obtain u1 where mra: "max_raise (av s, av t) = Some u1" by auto
  from max_raise_map_vars[OF gs gt mra[unfolded av_def]] obtain v1 where mr: "max_raise (s,t) = Some v1" and gv1: "ground v1" 
    and u1: "u1 = av v1" unfolding av_def by blast
  from r mra u1 have res: "max_raise_list (map av (v1 # ss)) = Some u" by simp
  from gv1 3(2) have "\<And> u. u \<in> set (v1 # ss) \<Longrightarrow> ground u" by auto
  from 3(1)[OF mr this res] show ?case using mr by auto
qed auto
  
definition e_bounded :: "(('f,'v)rule \<Rightarrow> ('f,'v)term \<Rightarrow> bool) \<Rightarrow> nat \<Rightarrow> ('f,'v)trs \<Rightarrow> ('f,'v)terms \<Rightarrow> bool"
  where "e_bounded e c R L \<equiv> \<forall> s t f n. s \<in> lift_term 0 ` L \<longrightarrow> (s,t) \<in> (rstep (cover e R))^* \<longrightarrow> (f,n) \<in> funas_term t \<longrightarrow> height f \<le> c"

definition e_raise_bounded :: "(('f,'v)rule \<Rightarrow> ('f,'v)term \<Rightarrow> bool) \<Rightarrow> nat \<Rightarrow> ('f,'v)trs \<Rightarrow> ('f,'v)terms \<Rightarrow> bool"
  where "e_raise_bounded e c R L \<equiv> \<forall> s t f n. s \<in> lift_term 0 ` L \<longrightarrow> (s,t) \<in> (raise_step (cover e R))^* \<longrightarrow> (f,n) \<in> funas_term t \<longrightarrow> height f \<le> c"

(* important direction of remark after Def. 11 in IC 09 *)
lemma left_bounded_raise_step_imp_rstep: assumes ll: "left_linear_trs R"
  shows "(s,t) \<in> raise_step R \<Longrightarrow> (s,t) \<in> rstep R"
proof (induct rule: raise_step.induct)
  case (raise_step l r C xs ss \<Theta> D)
  from raise_step(1) have lr: "(l,r) \<in> R" .
  from ll lr have l: "linear_term l" unfolding left_linear_trs_def by auto
  from split_vars_eqf_subst[of l \<Theta>, unfolded raise_step(2)] 
  have eqf: "l \<cdot> \<Theta> =f (C, map \<Theta> xs)" by auto
  from split_vars_vars_term_list'[of l, unfolded raise_step(2)] have xs: "xs = vars_term_list' l" by auto
  from raise_step(3) have len: "length ss = length xs" by auto
  from rstepI[OF lr refl refl] have "(D \<langle> l \<cdot> \<Theta> \<rangle>, D \<langle>r \<cdot> \<Theta> \<rangle>) \<in> rstep R" by auto
  also have "l \<cdot> \<Theta> = fill_holes C (map \<Theta> xs)" using eqfE[OF eqf] by auto
  also have "map \<Theta> xs = ss" 
  proof (rule nth_equalityI, unfold length_map, simp add: len, intro allI impI)
    fix i
    assume i: "i < length xs"    
    hence mem: "xs ! i \<in> set xs" by auto
    with xs have cond: "xs ! i \<in> vars_term l" by auto
    let ?fxs = "filter (op = (xs ! i)) xs"
    from linear_vars_term_list'[OF l, folded xs, of "xs ! i"]
    have "length ?fxs \<le> 1" .
    moreover from mem have "xs ! i \<in> set ?fxs" by auto
    ultimately have fxs: "?fxs = [xs ! i]" by (cases ?fxs, auto)
    let ?zip = "[(si, xi)\<leftarrow>zip ss xs . xi = xs ! i]"
    from i len have mem: "(ss ! i, xs ! i) \<in> set ?zip" unfolding set_filter set_zip by auto
    from i have "map \<Theta> xs ! i = \<Theta> (xs ! i)" by simp
    also have "\<dots> = the (max_raise_list (map fst ?zip))" unfolding raise_step(5) using cond by auto
    also have "?zip = [(ss ! i, xs ! i)]" 
    proof (cases ?zip)
      case Nil
      with mem show ?thesis by auto
    next
      case (Cons sx1 zs) note oCons = this
      show ?thesis
      proof (cases zs)
        case Nil
        with Cons mem show ?thesis by auto
      next
        case (Cons xs2 zs')
        def x \<equiv> "xs ! i"
        from arg_cong[OF oCons[unfolded Cons], of "\<lambda> l. length (map snd l)"] 
        have "2 \<le> length (map snd ?zip)" by auto
        also have "map snd ?zip = ?fxs" using len unfolding x_def[symmetric]
        proof (induct xs arbitrary: ss)
          case (Cons y ys sss)
          from Cons(2) obtain a ss where sss: "sss = a # ss" and len: "length ss = length ys" by (cases sss, auto)
          from Cons(1)[OF len] show ?case unfolding sss by auto
        qed auto
        finally show ?thesis unfolding fxs by simp
      qed
    qed 
    finally show "map \<Theta> xs ! i = ss ! i" by simp
  qed  
  finally show ?case .
qed

lemma e_bounded_left_linear_imp_raise_bounded: assumes bounded: "e_bounded e c R L" and ll: "left_linear_trs R"
  shows "e_raise_bounded e c R L" unfolding e_raise_bounded_def
proof (intro allI impI)
  fix s t f n
  assume L: "s \<in> lift_term 0 ` L"
  and steps: "(s,t) \<in> (raise_step (cover e R))^*"
  and fn: "(f,n) \<in> funas_term t"
  from left_bounded_raise_step_imp_rstep[OF cover_left_linear[OF ll, of e]]
  have subset: "raise_step (cover e R) \<subseteq> rstep (cover e R)" by auto
  show "height f \<le> c" 
    by (rule bounded[unfolded e_bounded_def, rule_format, OF L _ fn], insert steps rtrancl_mono[OF subset], auto)
qed

fun seq_gen :: "(('g,'v)term \<Rightarrow> ('f,'v)term) \<Rightarrow> ('f,'v)trs \<Rightarrow> (('f,'v)term \<Rightarrow> ('g,'v)term) \<Rightarrow> (nat \<Rightarrow> ('g,'v)term) \<Rightarrow> nat \<Rightarrow> ('f,'v)term" where 
  "seq_gen l R b ts 0 = l (ts 0)"
| "seq_gen l R b ts (Suc i) = (SOME t. (seq_gen l R b ts i, t) \<in> R \<and> b t = ts (Suc i))"

(* Thm 13. in IC 09, not instantiated to specific enrichments e *)
lemma e_raise_bounded: assumes var_cond: "\<And> l r. (l,r) \<in> R \<Longrightarrow> vars_term r \<subseteq> vars_term l"
  and lSN: "locally_terminating (cover e R \<union> raise)"
  and R: "finite R"
  and F: "finite F"
  and L: "\<Union>(funas_term ` L) \<subseteq> F"
  and bounded: "e_raise_bounded e c R L"
  shows "SN_on (rstep R) L"
proof 
  fix ts
  assume start: "ts 0 \<in> L" and ssteps: "\<forall> i. (ts i, ts (Suc i)) \<in> rstep R"
  hence steps: "\<And> i. (ts i, ts (Suc i)) \<in> rstep R" by auto
  let ?rel = "raise_step (cover e R)"
  let ?seq = "seq_gen (lift_term 0) ?rel base_term ts"
  def ss \<equiv> ?seq
  {
    fix i
    have "base_term (?seq i) = ts i \<and> (i > 0 \<longrightarrow> (?seq (i - 1), ?seq i) \<in> ?rel)"
    proof (induct i)
      case 0
      show ?case unfolding seq_gen.simps base_lift by auto
    next
      case (Suc i)
      def s \<equiv> "?seq i"
      from Suc have base: "base_term s = ts i" unfolding s_def by blast
      have id: "\<And> x. (0 < Suc i \<longrightarrow> x) = x" "Suc i - 1 = i" by auto
      let ?p = "\<lambda> t'. (s, t') \<in> ?rel \<and> base_term t' = ts (Suc i)"
      from rstep_imp_raise_step[rule_format, OF var_cond steps base, of e]
        obtain t' where *: "?p t'" by auto
      from someI[of ?p t', OF this] have "?p (SOME t'. ?p t')" .
      thus ?case unfolding seq_gen.simps id s_def[symmetric] by auto
    qed    
  } note steps = this[folded ss_def]  
  {
    fix i
    have "(ss i, ss (Suc i)) \<in> ?rel" "base_term (ss i) = ts i" using steps[of "Suc i"] steps[of i] by auto
  } note steps = this
  let ?G = "funas_trs R \<union> F"
  def GG \<equiv> ?G
  def G \<equiv> "{((f,l),n) | f l n. (f,n) \<in> ?G \<and> l \<le> c}"
  let ?rell = "(sig_step G (rstep raise))^* O sig_step G (rstep (cover e R))"
  let ?relll = "sig_step G (rstep (cover e R \<union> raise))"
  from R have "finite (funas_trs R)" by (rule finite_funas_trs)
  hence "finite (GG \<times> {l. l \<le> c})" using F unfolding GG_def by auto
  note fin = finite_imageI[OF this, of "\<lambda> ((f,n),l). ((f,l),n)"]
  have G: "finite G" unfolding G_def GG_def[symmetric] 
    by (rule HOL.subst[of _ _ finite, OF _ fin], force)
  {
    fix i 
    have steps': "(ss 0, ss i) \<in> ?rel^*"
    proof (induct i)
      case (Suc i)
      thus ?case using steps[of i] by auto
    qed auto
    have "ss 0 = lift_term 0 (ts 0)" unfolding ss_def by simp 
    with start have ss0: "ss 0 \<in> lift_term 0 ` L" by auto
    from bounded[unfolded e_raise_bounded_def, rule_format, OF ss0 steps']
    have limit: "\<And> f n. (f,n) \<in> funas_term (ss i) \<Longrightarrow> height f \<le> c" by auto
    {
      fix fhn
      assume mem: "fhn \<in> funas_term (ss i)"
      obtain f h n where f: "fhn = ((f,h),n)" by (cases fhn, auto)
      note mem = mem[unfolded f]
      have "fhn \<in> G" unfolding G_def
      proof(rule, intro exI conjI, rule f)
        from limit[OF mem] show "h \<le> c" by simp
      next
        from mem have "(f,n) \<in> funas_term (ts i)" 
          unfolding steps(2)[symmetric]
          unfolding funas_term_map_term by force
        also have "funas_term (ts i) \<subseteq> ?G"
        proof (induct i)
          case 0
          with start L show ?case by auto
        next
          case (Suc i)
          show ?case 
            by (rule rstep_preserves_funas_terms_var_cond[OF _ Suc `(ts i, ts (Suc i)) \<in> rstep R` var_cond], auto)
        qed
        finally show "(f,n) \<in> ?G" .
      qed
    }
    hence "funas_term (ss i) \<subseteq> G" by auto    
  } note limit = this 
  {
    fix i
    from steps(1)[of i] limit[of i] limit[of "Suc i"]
    have "(ss i, ss (Suc i)) \<in> sig_step G (raise_step (cover e R))" by auto
  } note steps = this
  {
    fix i
    have "(ss i, ss (Suc i)) \<in> ?rell"
      by (rule sig_raise_step_imp_sig_steps[OF _ steps], unfold G_def, auto)
    also have "?rell \<subseteq> (sig_step G (rstep (cover e R)) \<union> sig_step G (rstep raise))^+" by regexp
    also have "sig_step G (rstep (cover e R)) \<union> sig_step G (rstep raise) = ?relll"
    unfolding sig_step_def rstep_union by auto
    finally have "(ss i, ss (Suc i)) \<in> ?relll^+" .
  } note steps = this
  from lSN[unfolded locally_terminating_def, rule_format, OF G] 
  have SN: "SN ?relll" .
  hence "SN (?relll^+)" by (rule SN_imp_SN_trancl)
  with steps
  show False by auto
qed

hide_const seq_gen

lemma e_raise_match_bounded_linear_complexity_main: assumes wf: "wf_trs R" 
  and R: "finite R"
  and F: "finite F"
  and L: "\<Union>(funas_term ` L) \<subseteq> F"
  and non_duplicating: "\<And> l r. (l,r) \<in> R \<Longrightarrow> vars_term_ms r \<le> vars_term_ms l"
  and bounded: "e_raise_bounded match c R L"
  shows "\<exists> c. \<forall> n s t. s \<in> L \<longrightarrow> (s, t) \<in> (rstep R) ^^ n \<longrightarrow> n \<le> size s * c"
proof -
  from wf have var_cond: "\<And> l r. (l,r) \<in> R \<Longrightarrow> vars_term r \<subseteq> vars_term l"
    unfolding wf_trs_def by auto
  let ?G = "funas_trs R \<union> F"
  def GG \<equiv> ?G
  def G \<equiv> "{((f,l),n) | f l n. (f,n) \<in> ?G \<and> l \<le> c}"
  let ?rel = "raise_step (cover match R)"
  let ?rell = "sig_step G ?rel"
  let ?rel3 = "(sig_step G (rstep raise))^* O sig_step G (rstep (cover match R))"
  let ?rel4 = "sig_step G (rstep (cover match R \<union> raise))"
  {
    fix s t
    assume "(s,t) \<in> ?rell"
    from sig_raise_step_imp_sig_steps[of G _ _ "cover match R", OF _ this]
    have "(s,t) \<in> ?rel3" unfolding G_def by auto
    also have "?rel3 \<subseteq> ?rel4^* O ?rel4" unfolding sig_step_def rstep_union
      by (rule relcomp_mono[OF rtrancl_mono], auto)
    also have "\<dots> = ?rel4^+" by regexp
    finally have "(s,t) \<in> ?rel4^+" .
  }
  hence rel4: "?rell \<subseteq> ?rel4^+" by auto
  from R have "finite (funas_trs R)" by (rule finite_funas_trs)
  hence "finite (GG \<times> {l. l \<le> c})" using F unfolding GG_def by auto
  note fin = finite_imageI[OF this, of "\<lambda> ((f,n),l). ((f,l),n)"]
  have G: "finite G" unfolding G_def GG_def[symmetric] 
    by (rule HOL.subst[of _ _ finite, OF _ fin], force)
  from finite_list[OF R] obtain Rs where Rs: "R = set Rs" by auto
  let ?fs = "map (\<lambda> (l,r). length (funas_term_list' r)) Rs"
  def k \<equiv> "Suc (max_list ?fs)"
  have k2: "\<And> l r. (l,r) \<in> R \<Longrightarrow> length (funas_term_list' r) \<le> k" unfolding Rs k_def using max_list[of _ ?fs] by force
  have k1: "1 \<le> k" unfolding k_def by simp
  def cc \<equiv> "Suc c"
  have cc: "\<And> f n. (f,n) \<in> G \<Longrightarrow> height f < cc" unfolding cc_def G_def by force
  show ?thesis
  proof (rule exI[of _ "Suc k ^ cc"], intro allI impI)
    fix n s t 
    assume start: "s \<in> L" and st: "(s,t) \<in> (rstep R)^^n"
    from st have "\<exists> lt. (lift_term 0 s, lt) \<in> ?rell^^n \<and> base_term lt = t \<and> funas_term lt \<subseteq> G" 
    proof (induct n arbitrary: t)
      case 0
      hence id: "t = s" by simp
      have "funas_term (lift_term 0 s) \<subseteq> G" unfolding G_def using start L by (auto simp: funas_term_map_term)
      thus ?case
        by (intro exI[of _ "lift_term 0 s"], simp add: id base_lift, induct s, auto)
    next
      case (Suc n t)
      from Suc(2) obtain u where su: "(s,u) \<in> (rstep R)^^n" and ut: "(u,t) \<in> rstep R" by auto
      from Suc(1)[OF su] obtain lu where rrsteps: "(lift_term 0 s, lu) \<in> ?rell ^^ n" and base: "base_term lu = u" 
        and luG: "funas_term lu \<subseteq> G" by blast
      from rstep_imp_raise_step[rule_format, of R _ _ _ match, OF var_cond ut base]
      obtain lt where step: "(lu, lt) \<in> ?rel" and base: "base_term lt = t" by auto
      have steps: "(lift_term 0 s, lu) \<in> ?rel ^^ n"
        by (rule set_mp[OF relpow_mono rrsteps], auto)
      from steps step have rsteps: "(lift_term 0 s, lt) \<in> ?rel ^^ Suc n" by auto
      from start have s: "lift_term 0 s \<in> lift_term 0 ` L" by auto
      from bounded[unfolded e_raise_bounded_def, rule_format, OF s relpow_imp_rtrancl[OF rsteps]] 
      have height: "\<And> f n. (f,n) \<in> funas_term lt \<Longrightarrow> height f \<le> c" .
      have "funas_trs R \<subseteq> ?G" "funas_term s \<subseteq> ?G" using start L by auto
      from rsteps_preserve_funas_terms_var_cond[OF this relpow_imp_rtrancl[OF Suc(2)] var_cond]
      have G: "funas_term t \<subseteq> ?G" by auto
      {
        fix f k n
        assume *: "((f,k),n) \<in> funas_term lt"
        from height[OF *] have k: "k \<le> c" by simp
        from * have "(base (f,k),n) \<in> funas_term (base_term lt)" unfolding funas_term_map_term by force
        with base have "(f,n) \<in> funas_term t" by simp
        with k G have "((f,k),n) \<in> G" unfolding G_def by auto
      }
      hence ltG: "funas_term lt \<subseteq> G" by auto
      from step luG ltG have "(lu, lt) \<in> ?rell" by auto
      with rrsteps have "(lift_term 0 s, lt) \<in> ?rell^^Suc n" by auto
      with base ltG show ?case by blast
    qed
    then obtain lt where steps: "(lift_term 0 s, lt) \<in> ?rell^^n" by auto
    from trancl_steps_relpow[OF rel4 this] obtain m where m: "n \<le> m"
      and steps: "(lift_term 0 s, lt) \<in> ?rel4^^m" by auto
    from sig_match_union_raise_linear_bound[OF wf non_duplicating cc k2 k1 steps]
    have "m \<le> size s * Suc k ^ cc" by simp
    with m show "n \<le> size s * Suc k ^ cc" by simp
  qed
qed

lemma e_raise_match_bounded_linear_complexity: 
  fixes G :: "('f \<times> nat)list" and R :: "('f,'v)trs"
  assumes wf: "wf_trs R" 
  and R: "finite R"
  and F: "finite F"
  and L: "\<Union>(funas_term ` L) \<subseteq> F"
  and HL: "{s . funas_term s \<subseteq> H \<and> ground s} \<subseteq> L"
  and GH: "set G \<subseteq> H"
  and conH: "(con,0) \<in> H"
  and non_duplicating: "\<And> l r. (l,r) \<in> R \<Longrightarrow> vars_term_ms r \<le> vars_term_ms l"
  and bounded: "e_raise_bounded match c R L"
  shows "deriv_bound_measure_class (rstep R) (Derivational_Complexity G) (Comp_Poly 1)"
proof -
  from e_raise_match_bounded_linear_complexity_main[OF wf R F L non_duplicating bounded] obtain c 
  where lin: "\<And> n s t.  s \<in> L \<Longrightarrow> (s, t) \<in> rstep R ^^ n \<Longrightarrow> n \<le> size s * c" by blast
  show ?thesis unfolding deriv_bound_measure_class_def
  proof (rule, intro impI)
    fix n and t :: "('f,'v)term"
    assume "t \<in> terms_of_nat (Derivational_Complexity G) n"
    hence tG: "funas_term t \<subseteq> set G" and tn: "size t \<le> n" by auto
    have "deriv_bound (rstep R) t (n * c)" unfolding deriv_bound_def
    proof (rule, elim exE)
      fix s
      let ?\<sigma> = "(\<lambda> _. Fun con []) :: ('f,'v)subst"
      assume *: "(t, s) \<in> rstep R ^^ Suc (n * c)"
      have seq: "(t \<cdot> ?\<sigma>, s \<cdot> ?\<sigma>) \<in> rstep R ^^ Suc (n * c)"
        by (rule relpow_image[OF _ *], auto)
      from tG GH conH have "funas_term (t \<cdot> ?\<sigma>) \<subseteq> H" by (auto simp: funas_term_subst)
      moreover have "ground (t \<cdot> ?\<sigma>)" by simp
      ultimately have "t \<cdot> ?\<sigma> \<in> L" using HL by auto
      from lin[OF this seq] have "Suc (n * c) \<le> size t * c" by simp
      also have "\<dots> \<le> n * c" using tn by auto
      finally show False by simp
    qed
    thus "deriv_bound (rstep R) t (c * n ^ 1 + 0)" by (simp add: ac_simps)
  qed
qed

(* 
   getting e_bounded_ness criterion of Geser et. al via left-linearity 
   (adding raise to local termination does not harm, since known enrichments can cope with raise rules)
 *)
lemma e_bounded: assumes var_cond: "\<And> l r. (l,r) \<in> R \<Longrightarrow> vars_term r \<subseteq> vars_term l"
  and SN: "locally_terminating (cover e R \<union> raise)"
  and R: "finite R"
  and F: "finite F"
  and L: "\<Union>(funas_term ` L) \<subseteq> F"
  and ll: "left_linear_trs R"
  and bounded: "e_bounded e c R L"
  shows "SN_on (rstep R) L"
proof -
  from e_raise_bounded[OF var_cond SN R F L e_bounded_left_linear_imp_raise_bounded[OF bounded ll]]
  show ?thesis .
qed

hide_const (open) match
hide_const (open) roof
hide_const (open) raise

end

