(*
Copyright 2009-2013 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Simple_Termination
imports
  "$AFP/Well_Quasi_Orders/Wpo_Kruskal_Terms"
  "$AFP/Well_Quasi_Orders/Wpo_Extension"
  Signature_Extension
begin

interpretation "term": finite_tree Fun "fst \<circ> the \<circ> root" args
  by (unfold_locales) auto

abbreviation "emb \<equiv> term.term_hemb"
abbreviation "gterms \<equiv> term.terms"

inductive_set terms for F where
  "Var x \<in> terms F" |
  "(f, n) \<in> F \<Longrightarrow> length ts = n \<Longrightarrow> \<forall>t\<in>set ts. t \<in> terms F \<Longrightarrow> Fun f ts \<in> terms F"

lemma ground_funas_term_subseteq_gterms_conv:
  "ground t \<and> funas_term t \<subseteq> F \<longleftrightarrow> t \<in> gterms F"
proof
  assume "ground t \<and> funas_term t \<subseteq> F"
  then show "t \<in> gterms F"
    by (induct t) (auto, rule term.terms.intros, auto)
next
  assume "t \<in> gterms F"
  then show "ground t \<and> funas_term t \<subseteq> F"
    by (induct) auto
qed

lemma gtermsD: "t \<in> gterms F \<Longrightarrow> ground t \<and> funas_term t \<subseteq> F"
  by (metis ground_funas_term_subseteq_gterms_conv)

lemma terms_UNIV [simp]:
  "terms UNIV = (UNIV :: ('f, 'v) terms)"
proof
  show "(UNIV :: ('f, 'v) terms) \<subseteq> terms UNIV"
  proof
    fix t :: "('f, 'v) term"
    show "t \<in> terms UNIV"
      by (induct t) (auto intro: terms.intros)
  qed
qed auto

lemma wfp_on_gterms_with_const_imp_wfp_on:
  assumes "(c, 0) \<in> F"
  assumes "wfp_on P (gterms F)"
    and "\<And>s t \<sigma>. P s t \<Longrightarrow> P (s \<cdot> \<sigma>) (t \<cdot> \<sigma>)"
  shows "wfp_on P (terms F)"
proof (rule ccontr)
  assume "\<not> wfp_on P (terms F)"
  then obtain f where *: "\<And>i. f i \<in> terms F"
    and **: "\<And>i. P (f (Suc i)) (f i)" unfolding wfp_on_def by blast
  let ?\<sigma> = "\<lambda>_. Fun c []"
  let ?f = "\<lambda>i. f i \<cdot> ?\<sigma>"
  from ** and assms have "\<And>i. P (?f (Suc i)) (?f i)" by auto
  moreover have "\<And>i. ?f i \<in> gterms F"
  proof -
    fix i
    from * have "f i \<in> terms F" by auto
    then show "?f i \<in> gterms F" using `(c, 0) \<in> F`
    apply (induct)
    apply force
    apply auto
    apply (rule term.terms.intros)
    apply assumption
    apply simp_all
    done
  qed
  ultimately have "\<forall>i. ?f i \<in> gterms F \<and> P (?f (Suc i)) (?f i)" by blast
  then have "\<exists>g. \<forall>i. g i \<in> gterms F \<and> P (g (Suc i)) (g i)" by auto
  then show False using assms unfolding wfp_on_def by blast
qed

definition rewrite_relation where
  "rewrite_relation P = (
    (\<forall>s t C. P s t \<longrightarrow> P (C\<langle>s\<rangle>) (C\<langle>t\<rangle>)) \<and>
    (\<forall>s t \<sigma>. P s t \<longrightarrow> P (s \<cdot> \<sigma>) (t \<cdot> \<sigma>)))"

definition rewrite_order where
  "rewrite_order P = (rewrite_relation P \<and> po_on P UNIV)"

definition simplification_order_on where
  "simplification_order_on P F = (rewrite_order P \<and>
    (\<exists>Q. wpo_on Q F \<and> (\<forall>s t. emb F Q s t \<longrightarrow> P s t)))"

text {*A TRS is simply terminating if it is compatible with a simplification order.*}
definition "ST R \<longleftrightarrow> (\<exists>P. simplification_order_on P UNIV \<and> R \<subseteq> {(x, y). P y x})"

lemma gterms_imp_terms:
  fixes F :: "('a \<times> nat) set"
  shows "gterms F \<subseteq> terms F"
proof
  fix t :: "('a, 'b) term"
  assume "t \<in> gterms F"
  then show "t \<in> terms F"
    by (induct) (rule terms.intros, auto)
qed

lemma rev_chain_on_transp_on_less:
  assumes "\<And>i. f i \<in> A" and "\<And>i. P (f (Suc i)) (f i)"
    and "transp_on P A" and "i < j"
  shows "P (f j) (f i)"
proof -
  let ?P = "P\<inverse>\<inverse>"
  from assms have "chain_on ?P f A" by auto
  moreover
  from transp_on_converse [OF `transp_on P A`]
    have "transp_on ?P A" .
  ultimately
  show ?thesis
    using chain_on_transp_on_less [OF _ _ `i < j`] by blast
qed

lemma simplification_order_on_imp_wfp_on_gterms:
  assumes "simplification_order_on P F"
  shows "wfp_on P (gterms F)"
proof -
  from assms obtain Q
    where wpo: "wpo_on Q F"
    and po: "po_on P UNIV"
    and emb: "\<And>s t. emb F Q s t \<Longrightarrow> P s t"
    unfolding simplification_order_on_def rewrite_order_def
    by blast
  from term.wpo_on_terms [OF wpo]
    have "wpo_on (emb F Q) (gterms F)" .
  then have "po_on (emb F Q) (gterms F)"
    and "almost_full_on ((emb F Q)\<^sup>=\<^sup>=) (gterms F)" unfolding wpo_on_def by blast+
  from wf_and_no_antichain_imp_po_extension_wf [OF po_af_imp_wf_and_no_antichain [OF this]]
    show ?thesis using po_on_subset [OF _ po, simplified] and emb by blast
qed

lemma funas_term_terms_conv:
  "funas_term t \<subseteq> F \<longleftrightarrow> t \<in> terms F"
proof
  assume "funas_term t \<subseteq> F"
  then show "t \<in> terms F" by (induct t) (auto intro: terms.intros)
next
  assume "t \<in> terms F"
  then show "funas_term t \<subseteq> F" by (induct) auto
qed

lemma subtree_supt_conv:
  "term.subtree s t = t \<rhd> s"
proof
  assume "term.subtree s t"
  then show "t \<rhd> s" by (induct s t) auto
next
  assume "t \<rhd> s"
  then show "term.subtree s t" by (induct t s) auto
qed

lemma supt_imp_emb:
  assumes "s \<in> gterms F" and "s \<rhd> t"
  shows "emb F P t s"
  using assms
  unfolding subtree_supt_conv [symmetric]
  by (metis term.subtree_imp_term_hemb)

text {*Every simplification order is well-founded.*}
lemma simplification_order_on_imp_wfp_on:
  fixes P :: "('f, 'v) term \<Rightarrow> ('f, 'v) term \<Rightarrow> bool"
  assumes so: "simplification_order_on P UNIV"
  shows "wfp_on P UNIV"
proof (rule ccontr)
  from so obtain Q where ro: "rewrite_order P"
    and wpo: "wpo_on Q UNIV"
    and ext: "\<And>s t. emb UNIV Q s t \<Longrightarrow> P s t"
    unfolding simplification_order_on_def by blast
  then have po: "po_on P UNIV"
    and "rewrite_relation P" by (auto simp: rewrite_order_def)
  then have subst: "\<And>s t \<sigma>. P s t \<Longrightarrow> P (s \<cdot> \<sigma>) (t \<cdot> \<sigma>)" by (auto simp: rewrite_relation_def)
  from po have "transp_on P UNIV" by (rule po_on_imp_transp_on)
  then have trans: "transp_on P\<inverse>\<inverse> UNIV" by (metis transp_on_converse_simp)
  txt {*An arbitrary constant replaces all variables.*}
  obtain c::'f where "c \<in> UNIV" by blast
  txt {*The following is possible since substitutions are not restricted to finite domains.*}
  let ?\<sigma> = "(\<lambda>_. Fun c []) :: 'v \<Rightarrow> ('f, 'v) term"

  assume "\<not> ?thesis"
  then obtain f where *: "\<And>i. P (f (Suc i)) (f i)"
    by (auto simp: wfp_on_def)
  def T \<equiv> "\<lambda>i. f i \<cdot> ?\<sigma>"
  have "\<And>i. T i \<in> gterms UNIV"
  proof -
    fix i show "T i \<in> gterms UNIV" unfolding T_def
    proof (induct t\<equiv>"f i" arbitrary: f)
      case (Var x) show ?case by (auto simp: Var [symmetric])
    next
      case (Fun g ts)
      show ?case
        by (simp add: Fun(2) [symmetric], rule term.terms.intros)
           (insert Fun, auto)
    qed
  qed
  with term.almost_full_on_terms [OF wpo_on_imp_almost_full_on [OF wpo]]
    obtain i j where "i < j" and "(emb UNIV Q\<^sup>=\<^sup>=)\<^sup>=\<^sup>= (T i) (T j)"
    unfolding almost_full_on_def and good_def by blast
  with ext [of "T i" "T j"] have "P\<^sup>=\<^sup>= (T i) (T j)"
    by auto (metis (full_types) sup2E term.term_hemb_imp_terms term.term_hembeq_term_hemb_conv)
  moreover
  from chain_on_transp_on_less [of f UNIV "P\<inverse>\<inverse>", OF _ trans `i < j`] and *
    have "P (T j) (T i)" by (simp add: T_def subst)
  ultimately
  show False by (metis UNIV_I po po_on_imp_irreflp_on po_on_imp_transp_on strict_reflclp)
qed

lemma wfp_on_SN_conv:
  "wfp_on P UNIV \<longleftrightarrow> SN {(x, y). P y x}"
  unfolding wfp_on_def SN_defs by blast

lemma SN_wfp_on_conv:
  "SN R \<longleftrightarrow> wfp_on (\<lambda>x y. (y, x) \<in> R) UNIV"
  unfolding wfp_on_def SN_defs by blast

text {*Every simply terminating TRS is terminating.*}
corollary ST_imp_SN:
  assumes "ST R"
  shows "SN (rstep R)"
proof (rule ccontr)
  from assms obtain P where so: "simplification_order_on P UNIV"
    and *: "\<And>s t. (s, t) \<in> R \<Longrightarrow> P t s" unfolding ST_def by blast
  from so have ctxt: "\<And>s t C. P s t \<Longrightarrow> P C\<langle>s\<rangle> C\<langle>t\<rangle>"
    and subst: "\<And>s t \<sigma>. P s t \<Longrightarrow> P (s \<cdot> \<sigma>) (t \<cdot> \<sigma>)"
    by (auto simp: simplification_order_on_def rewrite_order_def rewrite_relation_def)
  from so and simplification_order_on_imp_wfp_on [of P]
    have sn: "SN {(x, y). P y x}" unfolding wfp_on_SN_conv by blast
  
  assume "\<not> ?thesis"
  then obtain f where **: "\<And>i. (f i, f (Suc i)) \<in> rstep R" by (auto simp: SN_defs)

  have "\<And>i. P (f (Suc i)) (f i)"
  proof -
    fix i
    from ** [of i] show "P (f (Suc i)) (f i)"
      by (induct) (auto dest: * subst ctxt)
  qed
  with sn show False by auto
qed

lemma wfp_on_terms_imp_SN_rstep:
  assumes subset: "funas_trs R \<subseteq> F"
    and compatible: "R \<subseteq> {(l, r). P r l}"
    and subst: "\<And>s t \<sigma>. P s t \<Longrightarrow> P (s \<cdot> \<sigma>) (t \<cdot> \<sigma>)"
    and ctxt: "\<And>s t C. P s t \<Longrightarrow> P (C\<langle>s\<rangle>) (C\<langle>t\<rangle>)"
    and "wfp_on P (terms F)"
  shows "SN (rstep R)"
proof (rule ccontr)
  let ?F = "funas_trs R"
  let ?R = "sig_step ?F (rstep R)"
  assume "\<not> SN (rstep R)"
  then have "\<not> SN ?R" unfolding SN_sig_rstep_SN_rstep_conv .
  then obtain f where *: "\<And>i. (f i, f (Suc i)) \<in> ?R" by (auto simp: SN_defs)
  have "\<forall>i. f i \<in> terms F \<and> P (f (Suc i)) (f i)"
  proof
    fix i
    from * have "funas_term (f i) \<subseteq> F" 
      using subset unfolding sig_step_def by blast
    then have "f i \<in> terms F" by (simp add: funas_term_terms_conv)
    moreover
    have "P (f (Suc i)) (f i)"
    proof -
      from * have "(f i, f (Suc i)) \<in> ?R" by simp
      then have "(f i, f (Suc i)) \<in> rstep R" unfolding sig_step_def by blast
      then show ?thesis
      proof
        fix l r C \<sigma>
        assume "(l, r) \<in> R" and fi: "f i = C\<langle>l \<cdot> \<sigma>\<rangle>" and fsi: "f (Suc i) = C\<langle>r \<cdot> \<sigma>\<rangle>"
        from `(l, r) \<in> R` and compatible have "P r l" by auto
        from ctxt [OF subst [OF this]] show ?thesis unfolding fi fsi by blast
      qed
    qed
    ultimately show "f i \<in> terms F \<and> P (f (Suc i)) (f i)" by blast
  qed
  with `wfp_on P (terms F)` show False unfolding wfp_on_def by blast
qed

text {*Compatibility with a simplification-order implies strong normalization
of a TRS.*}
lemma simplification_order_imp_SN_rstep:
  assumes so: "simplification_order_on P F"
    and "(c, 0) \<in> F"
    and "funas_trs R \<subseteq> F"
    and "R \<subseteq> {(l, r). P r l}"
  shows "SN (rstep R)"
proof (rule wfp_on_terms_imp_SN_rstep)
  show "funas_trs R \<subseteq> F" by fact
  show "R \<subseteq> {(l, r). P r l}" by fact
  from so show "\<And>s t \<sigma>. P s t \<Longrightarrow> P (s \<cdot> \<sigma>) (t \<cdot> \<sigma>)"
    and "\<And>s t C. P s t \<Longrightarrow> P C\<langle>s\<rangle> C\<langle>t\<rangle>"
    unfolding simplification_order_on_def rewrite_order_def rewrite_relation_def
    by auto
  show "wfp_on P (terms F)"
  proof (rule  wfp_on_gterms_with_const_imp_wfp_on)
    show "(c, 0) \<in> F" by fact
    show "\<And>s t \<sigma>. P s t \<Longrightarrow> P (s \<cdot> \<sigma>) (t \<cdot> \<sigma>)" by fact
    show "wfp_on P (gterms F)"
    proof (rule simplification_order_on_imp_wfp_on_gterms)
      show "simplification_order_on P F" by fact
    qed
  qed
qed

text {*Every well-founded relation can be extended to a well-partial-order.
TODO: apply well_order_extension lemma from the development version *}
lemma wfp_on_imp_wpo_on_extension:
  assumes "wfp_on P A"
  shows "\<exists>Q. (\<forall>x\<in>A. \<forall>y\<in>A. P x y \<longrightarrow> Q x y) \<and> wpo_on Q A"
  using wpo_on_extension [OF assms] .

end

